# -*- coding: utf-8 -*-
# @Time    : 2021/5/12 2:06 下午
# @Author  : shaoguowen
# @Email   : shaoguowen@tencent.com
# @FileName: censeo_ivqa_model.py
# @Software: PyCharm

import torch
from torch import nn

try:
  from torch.hub import load_state_dict_from_url
except ImportError:
  from torch.utils.model_zoo import load_url as load_state_dict_from_url

import tw

model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
    'resnext50_32x4d': 'https://download.pytorch.org/models/resnext50_32x4d-7cdf4587.pth',
    'resnext101_32x8d': 'https://download.pytorch.org/models/resnext101_32x8d-8ba56ff5.pth',
    'wide_resnet50_2': 'https://download.pytorch.org/models/wide_resnet50_2-95faca4d.pth',
    'wide_resnet101_2': 'https://download.pytorch.org/models/wide_resnet101_2-32ee1156.pth',
}


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
  """3x3 convolution with padding"""
  return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                   padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
  """1x1 convolution"""
  return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
  expansion = 1

  def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
               base_width=64, dilation=1, norm_layer=None):
    super(BasicBlock, self).__init__()
    if norm_layer is None:
      norm_layer = nn.BatchNorm2d
    if groups != 1 or base_width != 64:
      raise ValueError('BasicBlock only supports groups=1 and base_width=64')
    if dilation > 1:
      raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
    # Both self.conv1 and self.downsample layers downsample the input when stride != 1
    self.conv1 = conv3x3(inplanes, planes, stride)
    self.bn1 = norm_layer(planes)
    self.relu = nn.ReLU(inplace=True)
    self.conv2 = conv3x3(planes, planes)
    self.bn2 = norm_layer(planes)
    self.downsample = downsample
    self.stride = stride

  def forward(self, x):
    identity = x

    out = self.conv1(x)
    out = self.bn1(out)
    out = self.relu(out)

    out = self.conv2(out)
    out = self.bn2(out)

    if self.downsample is not None:
      identity = self.downsample(x)

    out += identity
    out = self.relu(out)

    return out


class Bottleneck(nn.Module):
  # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
  # while original implementation places the stride at the first 1x1 convolution(self.conv1)
  # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
  # This variant is also known as ResNet V1.5 and improves accuracy according to
  # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

  expansion = 4

  def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
               base_width=64, dilation=1, norm_layer=None):
    super(Bottleneck, self).__init__()
    if norm_layer is None:
      norm_layer = nn.BatchNorm2d
    width = int(planes * (base_width / 64.)) * groups
    # Both self.conv2 and self.downsample layers downsample the input when stride != 1
    self.conv1 = conv1x1(inplanes, width)
    self.bn1 = norm_layer(width)
    self.conv2 = conv3x3(width, width, stride, groups, dilation)
    self.bn2 = norm_layer(width)
    self.conv3 = conv1x1(width, planes * self.expansion)
    self.bn3 = norm_layer(planes * self.expansion)
    self.relu = nn.ReLU(inplace=True)
    self.downsample = downsample
    self.stride = stride

  def forward(self, x):
    identity = x

    out = self.conv1(x)
    out = self.bn1(out)
    out = self.relu(out)

    out = self.conv2(out)
    out = self.bn2(out)
    out = self.relu(out)

    out = self.conv3(out)
    out = self.bn3(out)

    if self.downsample is not None:
      identity = self.downsample(x)

    out += identity
    out = self.relu(out)

    return out


class ResNet(nn.Module):
  """
  ResNet, ref from pytorch official website
  """

  def __init__(self, block, layers, zero_init_residual=False,
               groups=1, width_per_group=64, replace_stride_with_dilation=None,
               norm_layer=None, input_channels=3, out_indices=(3,), strides=(2, 2, 2)):
    super(ResNet, self).__init__()
    if norm_layer is None:
      norm_layer = nn.BatchNorm2d
    self._norm_layer = norm_layer
    self.block = block
    self.inplanes = 64
    self.dilation = 1
    if replace_stride_with_dilation is None:
      # each element in the tuple indicates if we should replace
      # the 2x2 stride with a dilated convolution instead
      replace_stride_with_dilation = [False, False, False]
    if len(replace_stride_with_dilation) != 3:
      raise ValueError("replace_stride_with_dilation should be None "
                       "or a 3-element tuple, got {}".format(replace_stride_with_dilation))
    self.groups = groups
    self.base_width = width_per_group
    self.conv1 = nn.Conv2d(input_channels, self.inplanes, kernel_size=7, stride=2, padding=3,
                           bias=False)
    self.bn1 = norm_layer(self.inplanes)
    self.relu = nn.ReLU(inplace=True)
    self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
    self.ouput_dims = []
    self.layer1 = self._make_layer(block, 64, layers[0])
    self.ouput_dims.append(block.expansion * 64)
    self.layer2 = self._make_layer(block, 128, layers[1], stride=strides[0],
                                   dilate=replace_stride_with_dilation[0])
    self.ouput_dims.append(block.expansion * 128)
    self.layer3 = self._make_layer(block, 256, layers[2], stride=strides[1],
                                   dilate=replace_stride_with_dilation[1])
    self.ouput_dims.append(block.expansion * 256)
    self.layer4 = self._make_layer(block, 512, layers[3], stride=strides[2],
                                   dilate=replace_stride_with_dilation[2])
    self.ouput_dims.append(block.expansion * 512)
    self.out_indices = out_indices

    self.ouput_dims = [x for i, x in enumerate(self.ouput_dims) if i in out_indices]

    for m in self.modules():
      if isinstance(m, nn.Conv2d):
        nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
      elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
        nn.init.constant_(m.weight, 1)
        nn.init.constant_(m.bias, 0)

    # Zero-initialize the last BN in each residual branch,
    # so that the residual branch starts with zeros, and each residual block behaves like an identity.
    # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
    if zero_init_residual:
      for m in self.modules():
        if isinstance(m, Bottleneck):
          nn.init.constant_(m.bn3.weight, 0)
        elif isinstance(m, BasicBlock):
          nn.init.constant_(m.bn2.weight, 0)

  def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
    norm_layer = self._norm_layer
    downsample = None
    previous_dilation = self.dilation
    if dilate:
      self.dilation *= stride
      stride = 1
    if stride != 1 or self.inplanes != planes * block.expansion:
      downsample = nn.Sequential(
          conv1x1(self.inplanes, planes * block.expansion, stride),
          norm_layer(planes * block.expansion),
      )

    layers = []
    layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                        self.base_width, previous_dilation, norm_layer))
    self.inplanes = planes * block.expansion
    for _ in range(1, blocks):
      layers.append(block(self.inplanes, planes, groups=self.groups,
                          base_width=self.base_width, dilation=self.dilation,
                          norm_layer=norm_layer))

    return nn.Sequential(*layers)

  def forward(self, x):
    outs = []
    x = self.conv1(x)
    x = self.bn1(x)
    x = self.relu(x)
    x = self.maxpool(x)

    x1 = self.layer1(x)
    outs.append(x1)
    x2 = self.layer2(x1)
    outs.append(x2)
    x3 = self.layer3(x2)
    outs.append(x3)
    x4 = self.layer4(x3)
    outs.append(x4)

    outs = [x for i, x in enumerate(outs) if i in self.out_indices]
    return tuple(outs)


def _resnet(arch, block, layers, pretrained, progress, **kwargs):
  model = ResNet(block, layers, **kwargs)
  if pretrained:
    print('load pretrained model>>>>>')
    if isinstance(pretrained, str) and os.path.exists(pretrained):
      print("load pretrained from local:", pretrained)
      state_dict = torch.load(pretrained)
    else:
      print("load pretrained from url:", model_urls[arch])
      state_dict = load_state_dict_from_url(model_urls[arch],
                                            progress=progress)
    net_dict = model.state_dict()
    pretrained_dict = {k: v for k, v in state_dict.items() if
                       k in net_dict and v.size() == net_dict[k].size()}
    net_dict.update(pretrained_dict)
    model.load_state_dict(net_dict)
  return model


def resnet18(pretrained=False, progress=True, **kwargs):
  r"""ResNet-18 model from
  `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

  Args:
      pretrained (bool): If True, returns a model pre-trained on ImageNet
      progress (bool): If True, displays a progress bar of the download to stderr
  """
  return _resnet('resnet18', BasicBlock, [2, 2, 2, 2], pretrained, progress, **kwargs)


class SimpleHead(torch.nn.Module):
  """
  一个简单的模型预测头部
  """

  def __init__(self, feats_dims, **kwargs):
    super(SimpleHead, self).__init__()
    self.avg_pool = nn.AdaptiveAvgPool2d(1)
    out_num = kwargs.get("out_num", 1)
    self.fc_out = nn.Sequential(nn.Linear(feats_dims[-1], 1024),
                                nn.ReLU(inplace=True),
                                nn.Linear(1024, out_num))

  def forward(self, x):
    x = x[-1]
    x = self.avg_pool(x)
    x = torch.flatten(x, 1)
    x = self.fc_out(x)
    return x


class CenseoIVQAModel(nn.Module):
  """
  censeo的一个strong baseline model
  """

  def __init__(self, pretrained=None):
    super(CenseoIVQAModel, self).__init__()
    input_channels = 3
    self.backbone = resnet18(pretrained=True, out_indices=(3,), strides=(2, 2, 2), input_channels=input_channels)
    self.head = SimpleHead(feats_dims=self.backbone.ouput_dims, out_num=1)
    self.use_sigmoid = True

    if pretrained is None:
      tw.checkpoint.load_state_dict_from_url(
          self, '/cephFS/video_lab/checkpoints/quality_assess/CenseoQoE/model_ugc.pth')
    elif pretrained == 'ugc':
      tw.checkpoint.load_state_dict_from_url(
          self, '/cephFS/video_lab/checkpoints/quality_assess/CenseoQoE/model_ugc.pth')
    elif pretrained == 'pgc':
      tw.checkpoint.load_state_dict_from_url(
          self, '/cephFS/video_lab/checkpoints/quality_assess/CenseoQoE/model_pgc.pth')
    elif pretrained == 'game':
      tw.checkpoint.load_state_dict_from_url(
          self, '/cephFS/video_lab/checkpoints/quality_assess/CenseoQoE/model_game.pth')

  def preprocess(self, x):
    x = tw.transform.rgb_to_bgr(x)
    # x = x * 255.0
    h, w = x.shape[-2:]

    resize_h, resize_w = 1920, 1080
    sub_img_h, sub_img_w = 1280, 720
    flag = False
    if (w - h) * (resize_w - resize_h) < 0:
      flag = True
      resize_w, resize_h = resize_h, resize_w
      sub_img_w, sub_img_h = sub_img_h, sub_img_w
    w_scale = resize_w / w
    h_scale = resize_h / h
    scale = max(h_scale, w_scale)
    sub_img_h_, sub_img_w_ = int(h * scale), int(w * scale)

    x = tw.transform.resize(x, height=sub_img_h_, width=sub_img_w_, interpolation=tw.transform.RESIZE_MODE.BICUBIC)

    # x = cv2.resize(img, (sub_img_w_, sub_img_h_), cv2.INTER_CUBIC)
    h_i = (sub_img_h_ - sub_img_h) // 2
    w_i = (sub_img_w_ - sub_img_w) // 2
    x = x[..., h_i:h_i + sub_img_h, w_i:w_i + sub_img_w]
    if flag:
      x = torch.rot90(x, k=1, dims=(2, 3))

    return x

  def forward(self, x):
    # if self.config.fr:
    #   x_dst, x_ref = x.chunk(2, 2)
    #   x = torch.cat([x_dst, x_ref - x_dst], 2)
    if x.ndim == 4:
      x = x.unsqueeze(dim=1)

    n, t, c, h, w = x.size()
    x = x.view(n * t, c, h, w)
    feats = self.backbone(x)
    out = self.head(feats)
    if self.use_sigmoid:
      out = torch.sigmoid(out)
    out = out.view(n, t, -1)
    return out

  def forward_onnx(self, x):
    """
    专门给模型转换时使用
    :return:
    """
    # if self.config.fr:
    #   x_dst, x_ref = x.chunk(2, 1)
    #   x = torch.cat([x_dst, x_ref - x_dst], 1)
    feats = self.backbone(x)
    out = self.head(feats)
    if self.use_sigmoid:
      out = torch.sigmoid(out)
    return out
